#!/usr/bin/python3
# fn_data.py - 2021-01-29
# This library contains miscellaneous functions relating
# to types, strings, data, file operations, and maybe other
# things I've not thought of yet.

import inc
import objectifier
build_GG_command= objectifier.build_GG_command

# =============== Built In Functions ===========================
internal_syms= dict() # Establish the list of built-in functions.

def intf_NUM(E):
    """Converts a character to its ASCII value (e.g. 65-90 is A-Z). This is the
inverse of `chr`. It is called `ord` in Python. A text object that is a single
character will produce a single number. A text object of more than one
character will produce a list of numbers. A list of text objects is also valid
but will be considered as concatentated into one text object.
    ''A'' num -> 65.0
    ''AZ az'' num -> [65.0 90.0 32.0 97.0 122.0]
    [''xed'' ''.'' ''ch''] num -> [[120.0 101.0 100.0] 46.0 [99.0 104.0]]
"""
    # This should probably unpack SYMs too. But it doesn't.
    def pystr2num(s): # Input is len(1+) Python string. Output is stack ready num or list of them.
        if len(s) > 1: # More than one character to process; result is a list.
            oL= []
            for c in s:
                oL.append( objectifier.StackOB_VAL( ord(c) ) )
            return objectifier.StackOB_LST( oL )
        else: # Only one character; result is a single value.
            return objectifier.StackOB_VAL( ord(s) )
    o= None # If this doesn't change, then signals an error.
    v1= E.The.StackPop() # Input object, untouched and suitable for undo use.
    if v1.whatami == 'LST': # List input. Call this function on each element.
        oL= [] # Output list.
        for i in v1.val:
            if i.whatami == 'TXT' and len(i.val)>0: # Text input.
                oL.append( pystr2num(i.val) )
            else: # Error.
                oL= None 
                break # Let me stop you right there. This is not going to work.
        if oL is not None:
            o= objectifier.StackOB_LST( oL )
    else: # Not a list. Hopefully TXT.
        o= pystr2num(v1.val)
    if o is not None:
        E.The.StackPush(o)
    else: # If it's not a list and not text, then input is wrong.
        E.The.StackPush(v1) # Put the messed up input back and walk away.
        print("Input Error: num")
        print(intf_NUM.__doc__)
        return # Without doing much of anything.
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), v1 ] )
internal_syms['num']= intf_NUM # Register function.

def intf_CHR(E):
    """Converts an ASCII VAL to its character TXT (e.g. A-Z is 65-90). This is the
inverse of `num`. A single number will produce a single text character. A list
of numbers will produce a complete string. 
    65 chr -> ''A'' 
    [65 90 32 97 122] chr -> ''AZ az'' 
    [[120 101 100] 46 [99 104]] chr -> [''xed'' ''.'' ''ch'']
"""
    # This should probably unpack SYMs too. But it doesn't.
    def pylist2str(s): # Input is len(1+) Python list of ok nums. Output is stack ready TXT.
        if type(s) is type(list()): # More than one character to process; result is a list.
            oL= ''
            for c in s:
                oL+= chr(int(c.val))
            return oL
        else: # Only one character; result is a single value.
            return chr(int(s))
    o= None # If this doesn't change, then signals an error.
    v1= E.The.StackPop() # Input object, untouched and suitable for undo use.
    if v1.whatami == 'LST': # List input. Call this function on each element.
        oS= '' # Output string.
        for i in v1.val: # Hopefully a list of ok values. Or a list of lists thereof.
            if ((i.whatami == 'LST' and len(i.val)>0) # A sublist like each of these two [[1 2 3] [4 5 6]]
             or (i.whatami == 'VAL' and i.val>0)): # A proper positive value.
                oS= list(oS) # Make sure it's not the default empty string. Listing a list is a no-op.
                oS.append( objectifier.StackOB_TXT( pylist2str(i.val) ) )
            else: # Error.
                oS= None 
                break # Let me stop you right there. This is not going to work.
        if type(oS) is type(str()):
            o= objectifier.StackOB_TXT( oS )
        else: # Must be a list
            complex_check= [x.val for x in oS if len(x.val)>1]
            if complex_check != []: # This must be complicated like: [[99 99] 32 [65 66]]
                o= objectifier.StackOB_LST( oS )
            else: # Since this must be just a list of values, simply return one TXT object.
                o= objectifier.StackOB_TXT( ''.join([x.val for x in oS]) )
    else: # Not a list. Hopefully TXT.
        o= objectifier.StackOB_TXT( pylist2str(v1.val) )
    if o is not None:
        E.The.StackPush(o)
    else: # If it's not a list and not text, then input is wrong.
        E.The.StackPush(v1) # Put the messed up input back and walk away.
        print("Input Error: chr")
        print(intf_CHR.__doc__)
        return # Without doing much of anything.
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), v1 ] )
internal_syms['chr']= intf_CHR # Register function.
