#!/usr/bin/python
# fn_math.py $Revision: 1.3 $ $Date: 2014/04/11 04:45:53 $
# This library contains built-in math related functions.

import objectifier
import inc
# This library contains math related functions that all need the
# standard Python math library:
import math

build_GG_command= objectifier.build_GG_command

# =============== Built In Functions ===========================
internal_syms= dict() # Establish the list of built-in functions.


#COMPUTATIONAL
#-------------
# fabs(x) # name abs
def intf_ABS(E):
    """Absolute value."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    answer= objectifier.StackOB_VAL( math.fabs(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['abs']= intf_ABS # Register function.

# floor(x)
def intf_FLOOR(E):
    """Floor, lowest whole integer."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    answer= objectifier.StackOB_VAL( math.floor(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['floor']= intf_FLOOR # Register function.

# ceil(x)
def intf_CEIL(E):
    """Ceiling, highest whole integer."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    answer= objectifier.StackOB_VAL( math.ceil(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['ceil']= intf_CEIL # Register function.

# pow(x,y) # x to the y
def intf_POW(E):
    """Raises the value at level 2 to the value at level 1."""
    orig_num_1= E.The.StackPop() # Preserve this.
    orig_num_2= E.The.StackPop() # Preserve this.
    p= orig_num_1.val # Power part.
    b= orig_num_2.val # Base part.
    try:
        answer= objectifier.StackOB_VAL( math.pow(b,p) )
        E.The.StackPush(answer)
        E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'),
                                      orig_num_2, orig_num_1 ] )
    except ValueError:
        print('sqrt(-) - Second oldest trick in the book!')
        E.The.StackPush( orig_num_2 ) # Run away!
        E.The.StackPush( orig_num_1 ) # Run away!
internal_syms['pow']= intf_POW # Register function.

# XROOT - An HP48 synonym.
internal_syms['xroot']= build_GG_command('inv pow')
internal_syms['xroot'].__doc__= 'aka: inv pow'
internal_syms['xroot'].sd= '... OB2 OB1 xroot ==> ... OB1'

# sqrt(x)
def intf_SQRT(E):
    """Calculates the square root of level one."""
    X= E.The.StackPop() # Preserve this.
    x= X.val # Value of it.
    try:
        answer= objectifier.StackOB_VAL( math.sqrt(x) )
        E.The.StackPush(answer)
        E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
    except ValueError:
        print('sqrt(-) - Second oldest trick in the book!')
        E.The.StackPush( X ) # Run away!
internal_syms['sqrt']= intf_SQRT # Register function.

# SQ - An HP48 synonym.
internal_syms['sq']= build_GG_command('dup *')
internal_syms['sq'].__doc__= 'aka: dup *'
internal_syms['sq'].sd= '... OB1 sq ==> ... OB1'

# hypot(x,y)

# exp(x)

def intf_EXP(E):
    """Constant e to the value of level one.
        1 exp -> 2.718281828459045 
    Can be used on a list. Here is Softmax calculated:
        [1 2 3 4 1 2 3] exp dup sum / -> [.2036 .0643 .1747 .4748 .0236 .0643 .1747]"""
    if ( not inc.VAL(E) and
         not inc.LST_of_VALs(E) ):
        print("Input Error: exp")
        print(inc.VAL(E,1))
        print(" Value 1:"+str(E.The.StackPop()))
        print(intf_EXP.__doc__)
        return
    v1= E.resolve_symbol( E.The.StackPop() )
    answer= list()
    if v1.whatami == 'LST':
        for i in v1.val:
            if i.whatami == 'SYM':
                i= E.resolve_symbol(i)
            answer.append( objectifier.StackOB_VAL(math.exp(i.val)) )
        answerob= objectifier.StackOB_LST(answer)
        answerob.names= v1.names[:]
    else:
        #v1= E.The.StackPop() # Numeric input object.
        x= v1.val # Value of it.
        answerob= objectifier.StackOB_VAL( math.exp(x) )
    E.The.StackPush(answerob)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), v1 ] )
internal_syms['exp']= intf_EXP # Register function.

# log(x,b) # the base b logarithm of x
def intf_LOGBASE(E):
    """Logarithm of the value of level 2 with a base of the value of level 1."""
    orig_base=  E.The.StackPop() # Preserve in case of badness.
    orig_x=  E.The.StackPop() # Preserve in case of badness.
    b= orig_base.val
    x= orig_x.val
    try:
        E.The.StackPush(
             objectifier.StackOB_VAL( math.log(x,b) ) )
        E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), 
              orig_x, orig_base ] )
    except ZeroDivisionError:
        print('n/0 - Oldest trick in the book!')
        E.The.StackPush( orig_x ) # Run away!
        E.The.StackPush( orig_base ) # Run away!
internal_syms['logbase']= intf_LOGBASE # Register function.

# log(x) # the base e logarithm of x
def intf_LOG(E):
    """Base e logarithm of the value of level one. This is the natural logarithm
    which is the power to which e would have to be raised to equal the input.
    See also: logbase, log10, e_"""
    if ( not inc.VAL(E) and
         not inc.LST_of_VALs(E) ):
        print("Input Error: log")
        print(inc.VAL(E,1))
        print(" Value 1:"+str(E.The.StackPop()))
        print(intf_EXP.__doc__)
        return
    v1= E.resolve_symbol( E.The.StackPop() )
    answer= list()
    if v1.whatami == 'LST':
        for i in v1.val:
            if i.whatami == 'SYM':
                i= E.resolve_symbol(i)
            answer.append( objectifier.StackOB_VAL(math.log(i.val)) )
        answerob= objectifier.StackOB_LST(answer)
        answerob.names= v1.names[:]
    else:
        # This needs to be applied to the list form too. But really,
        # that form should use this sub function so only doing the check once.
        if v1.val <=0: 
            print("Input Error: non positive log")
            answerob= v1
        else:
            answerob= objectifier.StackOB_VAL( math.log(v1.val) )
    E.The.StackPush(answerob)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), v1 ] )
internal_syms['log']= intf_LOG # Register function.

# log10(x) # the base 10 logarithm of x
def intf_LOG10(E):
    """Base 10 logarithm of the value of level one. The power 10 would have to be raised
    to to equal the input."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    answer= objectifier.StackOB_VAL( math.log10(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['log10']= intf_LOG10 # Register function.

def intf_BASE(E):
    """\
Conversions between number base systems. Takes a number and a target base.
If the number object is a string, converts from the base system to decimal
equivalent whole number. If object is a value, converts to a string
representing it in specified base.
206 16 base  484 8 base  512 2 base -> ''CE'' ''744'' ''1000000000''
''bad'' 16 base  ''50'' 8 base  ''111'' 2 base -> 2989.0 40.0 7.0
    """
    if not (( inc.TXT(E,2) or inc.VAL(E,2) ) and ( inc.VAL(E,1) )):
        print("Input Error: base")
        print(intf_BASE.__doc__)
        return # Without doing much of anything.
    orig_base= E.The.StackPop()
    base= int(orig_base.val)
    orig_n= E.The.StackPop()
    if orig_n.whatami == 'VAL': # Decimal number to strange base string.
        n= int(orig_n.val)
        anint,o,s= type(int()),str(),str()
        if int(base) < 1:
            print("Undefined")
        if base == 1: o= "|"*n
        elif n == 0: o= "0"
        else:
            o= ''
            if n<0: s,n= '-',n*-1
            def digit_str(x):
                if x > 15: return '(%s)'%x
                if x > 9: return chr(55+x)
                return str(x)
            while n:
                o= digit_str(int(n%base))+o
                n= n//base
            o= s+o
        E.The.StackPush(objectifier.StackOB_TXT(o))
    elif orig_n.whatami == "TXT": # Strange base string to decimal.
        ns= orig_n.val
        if base>36:
            print("Input Error: can't understand notation from numbers in bases above 36")
            print(intf_BASE.__doc__)
            return # Without doing much of anything.
        E.The.StackPush( objectifier.StackOB_VAL( int(ns,base) ) )
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), orig_n, orig_base ] )
internal_syms['base']= intf_BASE # Register function.

# Set Angle Mode
# Actual use assuming degrees has proven to be too annoying when doing tasks
# like radians based homework. The HP48 allows one to change modes and so will
# GeoGad. This internal function sets a user accessible symbol that will keep
# track of preferences. The default which will be assumed when undefined is
# "degrees". This mode should also produce correct complex number behavior if
# that is implemented. Note that there are also commands "RECT", "CYLIN", and
# "SPHERE" to change the modes of 3-d coordinates.
def intf_DEGMODE(E):
    """\
Set the angle mode to "degree". This is stored in an accessible symbol '_gg_angmode'.
Trignometric functions will assume arguments in degrees.
Degree mode is the default when undefined."""
    angmode= objectifier.StackOB_TXT('degrees')
    E.symtab['_gg_angmode']= angmode
internal_syms['degmode']= intf_DEGMODE # Register function.

def intf_RADMODE(E):
    """\
Set the angle mode to "radian". Trignometric functions will assume
angle arguments in radians or produce angle results in radians. Degree
mode, *not* radian, is the default. This state is stored in an accessible
symbol, '_gg_angmode'. If that symbol is released with `unsto`, the mode
reverts to degrees."""
    angmode= objectifier.StackOB_TXT('radians')
    E.symtab['_gg_angmode']= angmode
internal_syms['radmode']= intf_RADMODE # Register function.

# Does anybody actually use gradian/gon/grad? Wikipedia says:
# "In the 1970s - 1990s, most scientific calculators offered the grad... for
# their trigonometric functions. In the 2010s, some scientific calculators lack
# support for gradians." Doesn't seem worth the obvious name clash potential.
#def intf_GRAD(E):
#    angmode= objectifier.StackOB_TXT('grad')
#    E.symtab['_gg_angmode']= angmode
#internal_syms['degmode']= intf_GRAD # Register function.

# Degrees(r)
def intf_2DEG(E):
    """Converts a radian angle value to degrees."""
    if ( not inc.VAL(E,1) and
         not inc.LST_of_VALs(E,1) ):
        print("Input Error: 2deg")
        print(inc.VAL(E,1))
        print(" Value 1:"+str(E.The.StackPop()))
        print(intf_2DEG.__doc__)
        return
    X= E.The.StackPop() # Numeric input object.
    if X.whatami == 'LST':
        L= X.val
        l= len(L)
        answer= 0 # Assume zero from empty list [] and other weird things.
        if l == 1: # Assume [degrees].
            answer= L[0].val
        elif l == 2: # Assume [degrees minutes].
            answer= L[0].val + L[1].val/60
        elif l > 2: # Assume [degrees minutes seconds maybe other stuff].
            answer= L[0].val + L[1].val/60 + L[2].val/3600
        answer= objectifier.StackOB_VAL( answer )
    else: # Not a list, must be a single normal value.
        r= X.val # Value of it.
        answer= objectifier.StackOB_VAL( math.degrees(r) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['2deg']= intf_2DEG # Register function.

# Radians(d)
def intf_2RAD(E):
    """Converts a degree angle value to radians."""
    if ( not inc.VAL(E,1) and
         not inc.LST_of_VALs(E,1) ):
        print("Input Error: 2rad")
        print(inc.VAL(E,1))
        print(" Value 1:"+str(E.The.StackPop()))
        print(intf_2RAD.__doc__)
        return
    X= E.The.StackPop() # Numeric input object.
    if X.whatami == 'LST':
        L= X.val
        l= len(L)
        if l == 0: # Assume zero from empty list [].
            answer= 0
        elif l == 1: # Assume [degrees].
            answer= L[0].val
        elif l == 2: # Assume [degrees minutes].
            answer= L[0].val + L[1].val/60
        elif l == 3: # Assume [degrees minutes seconds].
            answer= L[0].val + L[1].val/60 + L[2].val/3600
        answer= objectifier.StackOB_VAL( math.radians(answer) )
    else: # Not a list, must be a single normal value.
        d= X.val # Value of it.
        answer= objectifier.StackOB_VAL( math.radians(d) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['2rad']= intf_2RAD # Register function.
# Radians(d)
def intf_2RADx(E):
    """Converts a degree angle value to radians."""
    X= E.The.StackPop() # Numeric input object.
    d= X.val # Value of it.
    answer= objectifier.StackOB_VAL( math.radians(d) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
#internal_syms['2rad']= intf_2RAD # Register function.

# Degrees-Minutes-Arcseconds(d) -> [d m s]
def intf_2DMS(E):
    """\
Converts a degree angle value to a list of [degrees arcminutes arcseconds].
Note that this honors the angle mode so if `radmode` is set, the input is
assumed to be radians even though the output is degrees (like you asked for)."""
    if not inc.VAL(E,1):
        print("Input Error: 2dms")
        print(inc.VAL(E,1))
        print(" Value 1:"+str(E.The.StackPop()))
        print(intf_2DMS.__doc__)
        return # Without doing much of anything.
    X= E.The.StackPop() # Numeric input object.
    if '_gg_angmode' in E.symtab and E.symtab['_gg_angmode'].val == "radians":
        X= math.degrees(X.val)
    else:
        X= X.val
    de= int(X)
    mi= int((X-de)*60)
    se= X*3600 - de*3600 - mi*60
    deP= objectifier.StackOB_VAL( de )
    miP= objectifier.StackOB_VAL( mi )
    seP= objectifier.StackOB_VAL( se )
    fulldms= objectifier.StackOB_LST( [deP,miP,seP]  )
    fulldms.names= ['degrees','arcminutes','arcseconds']
    E.The.StackPush(fulldms)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['2dms']= intf_2DMS # Register function.

# sin(x)
def intf_SIN(E):
    """Sine of level one. Argument interpreted as degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
        x= math.radians(x)
    answer= objectifier.StackOB_VAL( math.sin(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['sin']= intf_SIN # Register function.

# cos(x)
def intf_COS(E):
    """Cosine of level one. Argument interpreted as degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
        x= math.radians(x)
    answer= objectifier.StackOB_VAL( math.cos(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['cos']= intf_COS # Register function.

# tan(x)
def intf_TAN(E):
    """Tangent of level one. Argument interpreted as degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
        x= math.radians(x)
    answer= objectifier.StackOB_VAL( math.tan(x) )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['tan']= intf_TAN # Register function.

# asin(x)
def intf_ASIN(E):
    """Arcsine of level one. Result in degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it. Note this is not an angle!
    try: #if x>1 or x<-1:
        answer= math.asin(x) # Radians
        if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
            answer= math.degrees(answer) # Now degrees if no _gg_angmode or it's not radians.
        answer= objectifier.StackOB_VAL( answer )
        E.The.StackPush(answer)
        E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
    except ValueError:
        print("Sine functions do not produce real values greater than 1 or less than -1.")
        E.The.StackPush( X ) # Run away!
internal_syms['asin']= intf_ASIN # Register function.

# acos(x)
def intf_ACOS(E):
    """Arccosine of level one. Result in degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it. Not an angle, no need for conversion.
    try: #if x>1 or x<-1:
        answer= math.acos(x) # Radians.
        if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
            answer= math.degrees(answer) # Now degrees if no _gg_angmode or it's not radians.
        answer= objectifier.StackOB_VAL( answer )
        E.The.StackPush(answer)
        E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
    except ValueError:
        print("Cosine functions do not produce real values greater than 1 or less than -1.")
        E.The.StackPush( X ) # Run away!
internal_syms['acos']= intf_ACOS # Register function.

# atan(x)
def intf_ATAN(E):
    """Arctangent of level one. Result in degrees by default. See "radmode" for radians."""
    X= E.The.StackPop() # Numeric input object.
    x= X.val # Value of it.
    answer= math.atan(x) # Radians.
    if '_gg_angmode' not in E.symtab or E.symtab['_gg_angmode'].val != "radians":
        answer= math.degrees(answer) # Now degrees if no _gg_angmode or it's not radians.
    answer= objectifier.StackOB_VAL( answer )
    E.The.StackPush(answer)
    E.Undo.StackPush( [ objectifier.StackOB_SYM('drop'), X ] )
internal_syms['atan']= intf_ATAN # Register function.

# atan2(y,x) # atan of y/x with signs considered. (Python math, not HP48)
# sinh(x) - (pow(e,x)-pow(e,-x))/2
# asinh(x) - ln(x+sqrt(x*x+1)) # Inverse hyperbolic sine.
# cosh(x) - (pow(e,x)+pow(e,-x))/2
# acosh(x) - ln(x+sqrt(x*x-1))  # Inverse hyperbolic cosine.
# tanh(x) = (pow(e,2*x)-1)/(pow(e,2*x)+1)
# atanh(x) = .5*ln((1+x)/(1-x)) # Inverse hyperbolic tangent. (Not HP48!)
